/*
 * This file is a part of DHCP client for TOPPERS/ASP w/ TINET
 * Copyright 2012-2013(C) by A.GOTANDA
 */

#include <kernel.h>
#include <sil.h>
#include <t_syslog.h>
#include "kernel_cfg.h"

#include <tinet_defs.h>
#include <tinet_config.h>

#include <net/if.h>
#include <net/ethernet.h>
#include <net/net.h>
#include <net/net_var.h>
#include <net/net_timer.h>

#include <netinet/in.h>
#include <netinet/in_var.h>
#include <netinet/in_itron.h>
#include <netinet/if_ether.h>

#include <net/ethernet.h>

#include <string.h>
#include <t_stdlib.h>

#include "tinet_cfg.h"
#include "dhcpc.h"


#if !defined(SUPPORT_INET4)
#error This component is for IPv4 only.
#endif

#if defined(SUPPORT_LOOP) || defined(SUPPORT_PPP)
#error This component is for Ethernet only.
#endif

#if defined(SUPPORT_TCP) && !defined(TCP_CFG_EXTENTIONS)
#error This component needs TCP_CFG_EXTENTIONS for using TCP.
#endif

#if !defined(ARP_CFG_CALLBACK_DUPLICATED) && !defined(DONT_CHECK_DUPLICATED)
#error This component needs ARP_CFG_CALLBACK_DUPLICATED to send DHCPDECLINE
#error You have to define DONT_CHECK_DUPLICATED if you do not want to this feature
#endif


#if defined(DHCPC_DEBUG)
#define SYSLOG(__l__)	syslog __l__
static char_t addr_str_mask[sizeof("255.255.255.255")];
static char_t addr_str_gw[sizeof("255.255.255.255")];
static char_t addr_str_ntp[sizeof("255.255.255.255")];
static char_t addr_str_dns[sizeof("255.255.255.255")];
static char_t addr_str_log[sizeof("255.255.255.255")];
static char_t addr_str_received[sizeof("255.255.255.255")];
static char_t addr_str_server[sizeof("255.255.255.255")];

static char_t addr_str_src[sizeof("255.255.255.255")];
#else
#define SYSLOG(__l__)
#endif

// DHCP task ưԤ֡TCP  task ưΤԤĤɬס
#define DHCPC_BOOTSTRAP_WAIT			 1000	/* [ms] */
// Packet ꤲƤʤä˺ФޤǤλ֤ν
#define	DHCPC_RESEND_INITIAL_INTERVAL		 4	/* [s] */
// ФƤʤä˺дֳ֤礹Ȥκ
#define	DHCPC_RESEND_MAX_INTERVAL		64	/* [s] */
// Packet л link  down ƤȤ retry Ԥ
#define DHCPC_LINKDOWN_WAIT			 2000	/* [ms] */


#define DHCP_PKT_OPT_LEN	64
/* DHCP packet */
typedef struct dhcp_pkt {
    uint8_t	op;		/* Message op code (1 = REQUEST, 2 = REPLY) */
    uint8_t	htype;		/* Hardware address type (1 = 10Mb ethernet) */
    uint8_t	hlen;		/* Hardware address length (6) */
    uint8_t	hops;		/* HOP count (client sets to zero) */
    uint32_t	xid;		/* Tracsaction ID
				   (Random number chosen by client) */
    uint16_t	secs;		/* Secs. elapsed since renewal started
				   (Filled in by client) */
    uint16_t	flags;		/* Flags
				   (bit 0: BROADCAST) */
    uint32_t	ciaddr;		/* Client IP address
				   Only filled by client in BOUND, RENEW
				   or REBINDING */
    uint32_t	yiaddr;		/* 'Your' (= client) IP address */
    uint32_t	siaddr;		/* Server's IP address
				   Returned in DHCPOFFER or DHCPACK */
    uint32_t	giaddr;		/* Relay agent IP address */
    uint8_t	chaddr[16];	/* Client hardware address */
    char	sname[64];	/* Server host name (ASCIZ) */
    char	file[128];	/* Boot file name (ASCIZ) */
    uint8_t	options[DHCP_PKT_OPT_LEN];	/* optional parameters field */
} T_DHCP_PKT;

enum dhcp_msg_type {
    DHCPDISCOVER = 1,
    DHCPOFFER,
    DHCPREQUEST,
    DHCPDECLINE,
    DHCPACK,
    DHCPNAK,
    DHCPRELEASE,
    DHCPINFORM
};

static T_IPEP		src_addr, dst_addr;
static UB		mac_addr[ETHER_ADDR_LEN];
static T_DHCP_PKT	dhcp_req_pkt, dhcp_resp_pkt;
static int		dhcp_state;

struct dhcp_got_addr_info dhcp_assigned = { 0, 0, 0 }, dhcp_received;

#define __DHCP_STATE_GEN_DATA
#include "dhcp_state.h"

#define	DHCP_FLG_RECEIVED	0x0001
#define	DHCP_FLG_T1_EXPIRED	0x0010
#define	DHCP_FLG_T2_EXPIRED	0x0020
#define	DHCP_FLG_LEASE_EXPIRED	0x0040
#define	DHCP_FLG_DUPLICATED	0x0080
#define	DHCP_FLG_LINKUP		0x0100
#define	DHCP_FLG_LINKDOWN	0x0200
#define	DHCP_FLG_EXPIRED	(DHCP_FLG_T1_EXPIRED		\
				 | DHCP_FLG_T2_EXPIRED		\
				 | DHCP_FLG_LEASE_EXPIRED	\
				 | DHCP_FLG_DUPLICATED)

#define	DHCP_FLG_RCV_NBLK_DONE	0x0001

static ID	tskid;

static uint32_t	tick = 0;

#if defined(DHCPC_CHECK_LINK_STATUS)
#include "h8300h_gcc/h8_3069f/h8_3069f.h"

#if !defined(H8P7DR_b)
#define	H8P7DR_b	H8P7DR
#endif

static volatile BOOL if_linkup_f = FALSE;

static BOOL
get_if_link_status(void)
{
    return ((*(uint8_t *)H8P7DR_b) & 0x80) == 0;
}
#else
static const BOOL if_linkup_f = TRUE;

static BOOL
get_if_link_status(void)
{
    return TRUE;
}
#endif


int
dhcpc_get_state(void)
{
    return dhcp_state;
}


const char *
dhcpc_get_state_name(void)
{
    struct dhcp_state_table *ptr;

    for (ptr = dhcp_state_table; ptr->name != NULL; ptr++) {
	if (ptr->id == dhcp_state) {
	    break;
	}
    }
    return ptr->name;
}


void
dhcpc_cyclic_handler(VP_INT exinf)
{
#if defined(DHCPC_CHECK_LINK_STATUS)
    static BOOL prev_if_link_status = 0;

    if_linkup_f = get_if_link_status();
    if (if_linkup_f != prev_if_link_status) {
	if (if_linkup_f) {
	    SYSLOG((LOG_NOTICE, "[DHCP:CYC] link up"));
	    iset_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_LINKUP);
	} else {
	    SYSLOG((LOG_NOTICE, "[DHCP:CYC] link down"));
	}
    }
    prev_if_link_status = if_linkup_f;
#endif

    ++tick;

    if (dhcp_assigned.lease_time != 0xffffffff) {
	if (0 < dhcp_assigned.lease_time) {
	    if (--dhcp_assigned.lease_time == 0) {
		memset(&dhcp_assigned, 0, sizeof(dhcp_assigned));
		in4_add_ifaddr(IPV4_ADDRANY, IPV4_ADDRANY);
		iset_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_LEASE_EXPIRED);
		iset_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_ADDR_EXPIRED);
	    }
/*
	    SYSLOG((LOG_NOTICE, "[DHCP:CYC,%d] Lease time = %d",
		   (ID)exinf, dhcp_assigned.lease_time));
*/
	}
	if (0 < dhcp_assigned.rebinding_time) {
	    if (--dhcp_assigned.rebinding_time == 0) {
		dhcp_assigned.renewal_time = 0;
		iset_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_T2_EXPIRED);
	    }
	}
	if (0 < dhcp_assigned.renewal_time) {
	    if (--dhcp_assigned.renewal_time == 0) {
		iset_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_T1_EXPIRED);
	    }
	}
    }

    /* iact_tsk(MSG_TASK); */
}


static int32_t
get_fuzz(int32_t secs)
{
    return (secs + (int32_t)((tick / 3) % 3)) - 1;
}


static ER_UINT	nblk_len;

ER
dhcpc_callback(ID cepid, FN fncd, VP p_parblk)
{
    ER err = E_OK;

    nblk_len = *(ER_UINT*)p_parblk;

    switch (fncd) {
      case TEV_UDP_RCV_DAT:
	err = udp_rcv_dat(cepid,
			  &src_addr, &dhcp_resp_pkt, sizeof(dhcp_resp_pkt),
			  TMO_FEVR);
	set_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_RECEIVED);
	SYSLOG((LOG_NOTICE, "[DHCP:%d] response received", tskid));
	break;
    }

    return err;
}

//------------------------------------------------------------------------------

// DHCPREQUEST packet ȤΩơDHCPDISCOVER Ѥν
void
dhcpc_make_req_pkt(T_DHCP_PKT *req)
{
    memset(req, 0x00, sizeof(T_DHCP_PKT));
    req->op = 1;
    req->htype = 1;
    req->hlen = 6;
    req->hops = 0;
    req->xid = htonl(0xdeadbeef);
    req->secs = 0;
    req->flags = 0;		/* not BROADCAST */
    req->ciaddr = 0;		/* Not yet assigned */
    req->yiaddr = 0;		/* Not yet assigned */
    req->siaddr = 0;		/* Sets by server */
    req->giaddr = 0;		/* Sets by DHCP relay agent */
    memcpy(req->chaddr, mac_addr, 6);
    req->file[0] = '\x00';
}


// DHCPREQUEST packet  option ꤹ
uint8_t *
dhcpc_set_req_opt(T_DHCP_PKT *req, int dhcp_msg_type)
{
    uint8_t	*ptr, *ptr2;
    uint16_t	max_pkt_len = htons(sizeof(T_DHCP_PKT));
    int		cnt, len;
    T_IN4_ADDR	tmp_addr;

    req->ciaddr = IPV4_ADDRANY;
    switch (dhcp_msg_type) {
      case DHCPRELEASE:
	req->ciaddr = dhcp_assigned.addr;
	break;
      case DHCPREQUEST:
	switch (dhcp_state) {
	  case DHCP_STATE_BOUND:
	  case DHCP_STATE_RENEWING:
	  case DHCP_STATE_REBINDING:
	    req->ciaddr = dhcp_assigned.addr;
	    break;
	}
	break;
    }

    ptr = req->options;
    *ptr++ = 0x63;	/* Magic cookie [0] */
    *ptr++ = 0x82;	/* Magic cookie [1] */
    *ptr++ = 0x53;	/* Magic cookie [2] */
    *ptr++ = 0x63;	/* Magic cookie [3] */
    *ptr++ = 53;	/* DHCP message type */
    *ptr++ = 1;		/*	(len = 1) */
    *ptr++ = (uint8_t)dhcp_msg_type;
    *ptr++ = 61;	/* Client-identifier */
    *ptr++ = 7;		/*	(len = 7) */
    *ptr++ = 1;		/*	(type = 1) */
    memcpy(ptr, mac_addr, 6);
    ptr += 6;

    tmp_addr = dhcp_assigned.addr;
    switch (dhcp_msg_type) {
      case DHCPREQUEST:
	if (dhcp_state == DHCP_STATE_RENEWING
	    || dhcp_state == DHCP_STATE_REBINDING) {
	    break;
	}
	tmp_addr = dhcp_received.addr;
	/* fall through */
      case DHCPDECLINE:
	*ptr++ = 50;	/* Requested-IP */
	*ptr++ = 4;	/*	(len = 4) */
	*ptr++ = tmp_addr >> 24;
	*ptr++ = tmp_addr >> 16;
	*ptr++ = tmp_addr >>  8;
	*ptr++ = tmp_addr >>  0;
	break;
    }

    tmp_addr = dhcp_assigned.server_id;
    switch (dhcp_msg_type) {
      case DHCPREQUEST:
	if (dhcp_state == DHCP_STATE_RENEWING
	    || dhcp_state == DHCP_STATE_REBINDING
	    || dhcp_state == DHCP_STATE_INITREBOOT) {
	    break;
	}
	tmp_addr = dhcp_received.server_id;
	/* fall through */
      case DHCPDECLINE:
      case DHCPRELEASE:
	*ptr++ = 54;	/* Server-ID */
	*ptr++ = 4;	/*	(len = 4) */
	*ptr++ = tmp_addr >> 24;
	*ptr++ = tmp_addr >> 16;
	*ptr++ = tmp_addr >>  8;
	*ptr++ = tmp_addr >>  0;
	break;
    }

    switch (dhcp_msg_type) {
      case DHCPDISCOVER:
      case DHCPINFORM:
      case DHCPREQUEST:
	*ptr++ = 55;	/* Parameter-Request */
	*ptr++ = 7;	/*	(len = 1) */
	*ptr++ = 1;	/*	(Subnet-Mask) */
	*ptr++ = 3;	/*	(Router) */
	*ptr++ = 4;	/*	(Time-Server) */
	*ptr++ = 6;	/*	(DNS) */
	*ptr++ = 7;	/*	(Log-Server) */
	*ptr++ = 58;	/*	(Renewal-Time) */
	*ptr++ = 59;	/*	(Rebinding-Time) */
	break;
    }
    switch (dhcp_msg_type) {
      case DHCPDISCOVER:
      case DHCPINFORM:
      case DHCPREQUEST:
	*ptr++ = 57;	/* Maximum DHCP Message Size */
	*ptr++ = 2;	/*	(len = 2) */
	*ptr++ = max_pkt_len >> 8;
	*ptr++ = max_pkt_len >> 0;
	break;
    }
    *ptr++ = 0xff;	/* End */
    len = DHCP_PKT_OPT_LEN - (ptr - req->options);
    for (ptr2 = ptr, cnt = 0; cnt < len; cnt++) {
	*ptr2++ = 0x00;
    }

    return ptr - 1;
}


ER
dhcpc_send_req_pkt(T_DHCP_PKT *req, int dhcp_msg_type)
{
    ER	err;

    dhcpc_set_req_opt(req, dhcp_msg_type);

    if (dhcp_state == DHCP_STATE_RENEWING) {
	dst_addr.ipaddr = dhcp_assigned.server_id;
    } else {
	dst_addr.ipaddr = IPV4_ADDR_BROADCAST;
    }

    if ((err = udp_snd_dat(UDP_DHCP_CEPID,
			   &dst_addr,
			   req, sizeof(T_DHCP_PKT),
			   TMO_FEVR)) < 0) {
	SYSLOG((LOG_NOTICE, "[DHCPC:%02d SND] error: %d, %s", UDP_DHCP_CEPID,
		err, itron_strerror(err)));
	return err;
    }
    return E_OK;
}


int
dhcpc_parse_resp_pkt(T_DHCP_PKT *resp, VP_INT exinf)
{
    uint8_t	*ptr, code, len;
    int		cnt;
    int		dhcp_msg_type = -1;

    memset(&dhcp_received, 0, sizeof(dhcp_received));

    ptr = resp->options;
    if (*ptr++ != 0x63 || *ptr++ != 0x82 || *ptr++ != 0x53 || *ptr++ != 0x63) {
	SYSLOG((LOG_NOTICE, "[DHCP:%d] Invalid response packet", tskid));
	return -1;
    }
    cnt = 0;
    while (cnt < 60) {
	++cnt;
	switch ((code = *ptr++)) {
	  case 0:	/* Pad */
	    break;
	  case 255:	/* End */
	    cnt = 60;
	    break;
	  default:
	    ++cnt;
	    len = *ptr++;
	    switch (code) {
	      case 1:	// Subnet-Mask
		dhcp_received.mask
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   Subnet-Mask = %s",
			tskid, ip2str(addr_str_mask, &dhcp_received.mask)));
		break;
	      case 3:	// Router
		dhcp_received.gw
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   Default GW = %s",
			tskid, ip2str(addr_str_gw, &dhcp_received.gw)));
		break;
	      case 4:	// Time-Server
		dhcp_received.ntps
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   Time Server = %s",
			tskid, ip2str(addr_str_ntp, &dhcp_received.ntps)));
		break;
	      case 6:	// Domain-Name-Server
		dhcp_received.dns
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   Domain Name Server = %s",
			tskid, ip2str(addr_str_dns, &dhcp_received.dns)));
		break;
	      case 7:	// Log-Server
		dhcp_received.log
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   Log Server = %s",
			tskid, ip2str(addr_str_log, &dhcp_received.log)));
		break;
	      case 51:	// Lease-Time
		dhcp_received.lease_time = ptr[0];
		dhcp_received.lease_time <<= 8;
		dhcp_received.lease_time += ptr[1];
		dhcp_received.lease_time <<= 8;
		dhcp_received.lease_time += ptr[2];
		dhcp_received.lease_time <<= 8;
		dhcp_received.lease_time += ptr[3];
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   lease time = %d",
			tskid, dhcp_received.lease_time));
		break;
	      case 53:	// DHCP-Message
		dhcp_msg_type = *ptr;
		dhcp_received.addr = resp->yiaddr;
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   %s address: %s",
			tskid,
			dhcp_msg_type == DHCPOFFER ? "Offered" :
			dhcp_msg_type == DHCPACK ? "Assigned" : "",
			ip2str(addr_str_received, &dhcp_received.addr)));
		break;
	      case 54:	// Server-ID
		dhcp_received.server_id
		    = MAKE_IPV4_ADDR(ptr[0], ptr[1], ptr[2], ptr[3]);
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   server-ID = %s",
			tskid,
			ip2str(addr_str_server, &dhcp_received.server_id)));
		break;
	      case 58:	// Renewal-Time
		dhcp_received.renewal_time = ptr[0];
		dhcp_received.renewal_time <<= 8;
		dhcp_received.renewal_time += ptr[1];
		dhcp_received.renewal_time <<= 8;
		dhcp_received.renewal_time += ptr[2];
		dhcp_received.renewal_time <<= 8;
		dhcp_received.renewal_time += ptr[3];
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   renewal time = %d",
			tskid, dhcp_received.renewal_time));
		break;
	      case 59:	// Rebinding-Time
		dhcp_received.rebinding_time = ptr[0];
		dhcp_received.rebinding_time <<= 8;
		dhcp_received.rebinding_time += ptr[1];
		dhcp_received.rebinding_time <<= 8;
		dhcp_received.rebinding_time += ptr[2];
		dhcp_received.rebinding_time <<= 8;
		dhcp_received.rebinding_time += ptr[3];
		SYSLOG((LOG_NOTICE, "[DHCP:%d]   rebinding time = %d",
			tskid, dhcp_received.rebinding_time));
		break;
	    }
	    cnt += len;
	    ptr += len;
	    break;
	}
    }

    /* Set default netmask if not provided */
    if (dhcp_received.mask == 0 && dhcp_received.addr != 0) {
	if ((dhcp_received.addr & 0x80000000) == 0) {
	    /* Class A default netmask */
	    dhcp_received.mask = 0xff000000;
	} else if ((dhcp_received.addr & 0xc0000000) == 0x80000000) {
	    /* Class B default netmask */
	    dhcp_received.mask = 0xffff0000;
	} else if ((dhcp_received.addr & 0xe0000000) == 0xc0000000) {
	    /* Class C default netmask */
	    dhcp_received.mask = 0xffffff00;
	}
    }

    if (dhcp_msg_type == DHCPACK) {
	dhcp_assigned = dhcp_received;

	if (dhcp_assigned.renewal_time == 0) {
	    dhcp_assigned.renewal_time = dhcp_assigned.lease_time / 2;
	}
	if (dhcp_assigned.rebinding_time == 0) {
	    dhcp_assigned.rebinding_time
		= (uint32_t)(dhcp_assigned.lease_time * 0.875);
	}
    }

    return dhcp_msg_type;
}


//ȥ꡼ݥ
void
dhcpc_MainTask(VP_INT exinf)
{
    T_IF_SOFTC	*ic;
    ER		err;
#ifdef UDP_CFG_EXTENTIONS
    T_UDP_CCEP	ccep;
#endif
    FLGPTN	flgptn;
    int		dhcp_msg_type = -1;
    TMO		resend_interval;
    TMO		tmout;

    get_tid(&tskid);
    SYSLOG((LOG_NOTICE, "[DHCP:%d] started.", tskid));

    // ϥɥγ
    tick = 0;
    sta_cyc(DHCPC_CYCHDR1);

    // ֤֤UDP ưΤԤ
    // ʤʤ MAC address Ǥʤ
    tslp_tsk(DHCPC_BOOTSTRAP_WAIT);

    // MAC address 
    ic = IF_ETHER_NIC_GET_SOFTC();
    memcpy(mac_addr, ic->ifaddr.lladdr, ETHER_ADDR_LEN);
    SYSLOG((LOG_NOTICE, "[DHCP:%d] MAC address = %s",
	    tskid, mac2str(NULL, ic->ifaddr.lladdr)));

    // 
    dst_addr.ipaddr = IPV4_ADDR_BROADCAST;	// Packet ˾񤭤
    dst_addr.portno = DHCPS_PORTNO;		// ˸

#ifdef UDP_CFG_EXTENTIONS
    // NIC ν
    in4_add_ifaddr(IPV4_ADDRANY, IPV4_ADDRANY);

    ccep.cepatr = UINT_C(0);
    ccep.myaddr.ipaddr = IPV4_ADDRANY;
    ccep.myaddr.portno = DHCPC_PORTNO;
    ccep.callback = (FP)dhcpc_callback;
    if ((err = udp_cre_cep(UDP_DHCP_CEPID, &ccep)) != E_OK) {
	SYSLOG((LOG_NOTICE, "[DHCPC:%02d TSK] CEP create error: %s",
		UDP_DHCP_CEPID, itron_strerror(err)));
	ext_ker();
    }
#endif	/* of #ifdef UDP_CFG_EXTENTIONS */

    dhcp_assigned.lease_time = 0;
    dhcp_assigned.renewal_time = 0;
    dhcp_assigned.rebinding_time = 0;
    dhcp_assigned.server_id = IPV4_ADDRANY;
    dhcp_assigned.addr = dhcp_assigned.mask = IPV4_ADDRANY;
    dhcp_assigned.gw = IPV4_ADDRANY;
    dhcp_assigned.dns = dhcp_assigned.ntps = IPV4_ADDRANY;

    // DHCPREQUEST packet ȤΩ
    dhcpc_make_req_pkt(&dhcp_req_pkt);

    dhcp_state = DHCP_STATE_INIT;
    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
    while (1) {
	if (!get_if_link_status()) {
	    // I/F  link Ƥ link up  lease ڤޤԵ
/*
	    SYSLOG((LOG_NOTICE, "[DHCP:%d] Link is down", tskid));
*/
	    wai_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_LINKUP | DHCP_FLG_EXPIRED,
		    TWF_ORW, &flgptn);
	}

	if (get_if_link_status()) {
	    // Link is up
	    switch (dhcp_state) {
	      case DHCP_STATE_INIT:
		in4_add_ifaddr(IPV4_ADDRANY, IPV4_ADDRANY);
		clr_flg(DHCPC_MAIN_FLAGS, ~DHCP_FLG_RECEIVED);

		SYSLOG((LOG_NOTICE,
			"[DHCP:%d] Sending DHCPDISCOVER...", tskid));
		if (dhcpc_send_req_pkt(&dhcp_req_pkt, DHCPDISCOVER) != E_OK) {
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    if (DHCPC_RESEND_MAX_INTERVAL < (resend_interval *= 2)) {
			resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    }
		    continue;
		}
		dhcp_state = DHCP_STATE_SELECTING;
		set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		/* fall through */
	      case DHCP_STATE_SELECTING:
		tmout = get_fuzz(resend_interval) * 1000;
		SYSLOG((LOG_NOTICE,
			"[DHCP:%d] Waiting DHCPOFFER... (tmout = %d)",
			tskid, tmout));
		if (twai_flg(DHCPC_MAIN_FLAGS,
			     DHCP_FLG_RECEIVED | DHCP_FLG_EXPIRED,
			     TWF_ORW, &flgptn, tmout) == E_TMOUT) {
		    dhcp_state = DHCP_STATE_INIT;
		    if (DHCPC_RESEND_MAX_INTERVAL < (resend_interval *= 2)) {
			resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    }
		    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		    continue;
		}
		if (flgptn & DHCP_FLG_EXPIRED) {
		    break;
		}
		SYSLOG((LOG_NOTICE, "[DHCP:%d] Received DHCPOFFER from %s",
			tskid, ip2str(addr_str_src, &src_addr.ipaddr)));
		if ((dhcp_msg_type = dhcpc_parse_resp_pkt(&dhcp_resp_pkt,
							  exinf)) < 0) {
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    dhcp_state = DHCP_STATE_INIT;
		    if (DHCPC_RESEND_MAX_INTERVAL < (resend_interval *= 2)) {
			resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    }
		    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		    continue;
		}
		if (dhcp_msg_type != DHCPOFFER) {
		    SYSLOG((LOG_NOTICE,
			    "[DHCP:%d] Packet received, but not DHCPOFFER",
			    tskid));
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    dhcp_state = DHCP_STATE_INIT;
		    if (DHCPC_RESEND_MAX_INTERVAL < (resend_interval *= 2)) {
			resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    }
		    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		    continue;
		}
		dhcp_state = DHCP_STATE_REQUESTING;
		resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		/* fall through */
	      case DHCP_STATE_REQUESTING:
	      case DHCP_STATE_RENEWING:
	      case DHCP_STATE_REBINDING:
	      case DHCP_STATE_INITREBOOT:
		clr_flg(DHCPC_MAIN_FLAGS, ~DHCP_FLG_RECEIVED);

		SYSLOG((LOG_NOTICE, "[DHCP:%d] Sending DHCPREQUEST...", tskid));
		if ((err = dhcpc_send_req_pkt(&dhcp_req_pkt,
					      DHCPREQUEST)) != E_OK) {
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] Send error", tskid));
		    dhcp_state = DHCP_STATE_INIT;
		    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    continue;
		}

#if defined(UDP_CFG_EXTENTIONS) && defined(UDP_SOCK_REOPEN)
		SYSLOG((LOG_NOTICE, "[DHCPC:%02d] CEP delete", UDP_DHCP_CEPID));
		if ((err = udp_del_cep(UDP_DHCP_CEPID)) != E_OK) {
		    SYSLOG((LOG_NOTICE, "[DHCPC:%02d] CEP delete error: %s",
			    UDP_DHCP_CEPID, itron_strerror(err)));
		    ext_ker();
		}
#endif	/* of #if defined(UDP_CFG_EXTENTIONS) && defined(UDP_SOCK_REOPEN) */
		in4_add_ifaddr(dhcp_received.addr, dhcp_received.mask);
#if defined(UDP_CFG_EXTENTIONS) && defined(UDP_SOCK_REOPEN)
		ccep.myaddr.ipaddr = dhcp_received.addr;
		SYSLOG((LOG_NOTICE, "[DHCPC:%02d] CEP create", UDP_DHCP_CEPID));
		if ((err = udp_cre_cep(UDP_DHCP_CEPID, &ccep)) != E_OK) {
		    SYSLOG((LOG_NOTICE, "[DHCPC:%02d] CEP create error: %s",
			    UDP_DHCP_CEPID, itron_strerror(err)));
		    ext_ker();
		}
#endif	/* of #if defined(UDP_CFG_EXTENTIONS) && defined(UDP_SOCK_REOPEN) */

		tmout = get_fuzz(resend_interval) * 1000;
		SYSLOG((LOG_NOTICE,
			"[DHCP:%d] Waiting DHCP Response... (tmout = %d)",
			tskid, tmout));
		if (twai_flg(DHCPC_MAIN_FLAGS,
			     DHCP_FLG_RECEIVED | DHCP_FLG_EXPIRED,
			     TWF_ORW, &flgptn, tmout) == E_TMOUT) {
		    switch (dhcp_state) {
		      case DHCP_STATE_RENEWING:
			if ((resend_interval = dhcp_assigned.rebinding_time / 2) < 60) {
			    resend_interval = 60;
			}
			break;
		      case DHCP_STATE_REBINDING:
			if ((resend_interval = dhcp_assigned.lease_time / 2) < 60) {
			    resend_interval = 60;
			}
			break;
		      default:	/* maybe DHCP_STATE_REQUESTING */
			if (DHCPC_RESEND_MAX_INTERVAL < (resend_interval *= 2)) {
			    dhcp_state = DHCP_STATE_INIT;
			    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
			    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
			}
			break;
		    }
		    SYSLOG((LOG_NOTICE,
			    "[DHCP:%d] Timed out, next interval = %d",
			    tskid, resend_interval));
		    continue;
		}
		if (flgptn & DHCP_FLG_EXPIRED) {
		    break;
		}
		SYSLOG((LOG_NOTICE, "[DHCP:%d] Received DHCP Response from %s",
			tskid, ip2str(addr_str_src, &src_addr.ipaddr)));
		if ((dhcp_msg_type = dhcpc_parse_resp_pkt(&dhcp_resp_pkt,
							  exinf)) < 0) {
		    dhcp_state = DHCP_STATE_INIT;
		    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    set_flg(DHCPC_STATE_CHG_FLAGS,
			    DHCPC_FLG_ADDR_EXPIRED | DHCPC_FLG_STATE_CHANGED);
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    continue;
		}
		if (dhcp_msg_type == DHCPNAK) {
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] DHCPNAK received", tskid));
		    dhcp_state = DHCP_STATE_INIT;
		    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    set_flg(DHCPC_STATE_CHG_FLAGS,
			    DHCPC_FLG_ADDR_EXPIRED | DHCPC_FLG_STATE_CHANGED);
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    continue;
		}
		if (dhcp_msg_type != DHCPACK) {
		    SYSLOG((LOG_NOTICE,
			    "[DHCP:%d] Packet received, but not DHCPACK", tskid));
		    dhcp_state = DHCP_STATE_INIT;
		    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
		    set_flg(DHCPC_STATE_CHG_FLAGS,
			    DHCPC_FLG_ADDR_EXPIRED | DHCPC_FLG_STATE_CHANGED);
		    tmout = get_fuzz(resend_interval) * 1000;
		    SYSLOG((LOG_NOTICE, "[DHCP:%d] sleep %d",
			    tskid, resend_interval));
		    tslp_tsk(tmout);
		    continue;
		}
		dhcp_state = DHCP_STATE_BOUND;
		set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
		/* fall through */
	      case DHCP_STATE_BOUND:
		SYSLOG((LOG_NOTICE, "[DHCP:%d] Bound to %s",
			tskid, ip2str(NULL, &dhcp_assigned.addr)));
		if (dhcp_assigned.lease_time == 0xffffffff) {
		    SYSLOG((LOG_NOTICE, "[DHCP:%d]   Lease time = infinity",
			    tskid));
		} else {
		    SYSLOG((LOG_NOTICE,
			    "[DHCP:%d]   Lease time = %d / T2 = %d / T1 = %d",
			    tskid,
			    dhcp_assigned.lease_time,
			    dhcp_assigned.rebinding_time,
			    dhcp_assigned.renewal_time));
		}

		// ʣȸͤ gratuitous ARP 
		arp_request(&ic->ifaddr, dhcp_assigned.addr);

		// Address λ flag  set
		set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_ADDR_ASSIGNED);

		// ϤԤ
		wai_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_EXPIRED, TWF_ORW, &flgptn);
		break;
	    }
	}

	if (flgptn & DHCP_FLG_DUPLICATED) {
	    SYSLOG((LOG_NOTICE, "[DHCP:%d] Sending DHCPDECLINE...", tskid));
	    dhcpc_send_req_pkt(&dhcp_req_pkt, DHCPDECLINE);
	    dhcp_state = DHCP_STATE_INIT;
	    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
	    set_flg(DHCPC_STATE_CHG_FLAGS,
		    (DHCPC_FLG_ADDR_EXPIRED | DHCPC_FLG_STATE_CHANGED));
	} else if (flgptn & DHCP_FLG_LEASE_EXPIRED) {
	    SYSLOG((LOG_NOTICE, "[DHCP:%d] Lease time expired", tskid));
	    dhcp_state = DHCP_STATE_INIT;
	    resend_interval = DHCPC_RESEND_INITIAL_INTERVAL;
	    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
	} else if (flgptn & DHCP_FLG_T2_EXPIRED) {
	    SYSLOG((LOG_NOTICE, "[DHCP:%d] T2 expired", tskid));
	    dhcp_state = DHCP_STATE_REBINDING;
	    if ((resend_interval = dhcp_assigned.lease_time / 2) < 60) {
		resend_interval = 60;
	    }
	    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
	} else if (flgptn & DHCP_FLG_T1_EXPIRED) {
	    SYSLOG((LOG_NOTICE, "[DHCP:%d] T1 expired", tskid));
	    dhcp_state = DHCP_STATE_RENEWING;
	    if ((resend_interval = dhcp_assigned.rebinding_time / 2) < 60) {
		resend_interval = 60;
	    }
	    set_flg(DHCPC_STATE_CHG_FLAGS, DHCPC_FLG_STATE_CHANGED);
	}
    }

#if 0
    /*
     * DHCPRELEASE 
     */
    SYSLOG((LOG_NOTICE, "[DHCP:%d] Sending DHCPRELEASE...", tskid));
    dhcpc_send_req_pkt(&dhcp_req_pkt, DHCPRELEASE);
    // DHCPRELEASE  response Ϥʤ
    SYSLOG((LOG_NOTICE, "[DHCP:%d] Done.", tskid));

    ext_ker();
#endif
}


bool_t
arp_callback_duplicated(uint8_t *shost)
{
    static uint8_t shost_addr[sizeof("00:00:00:00:00:00 ")];

    mac2str(shost_addr, shost);
    SYSLOG((LOG_NOTICE,
	    "[DHCP:%d] Assigned address is duplicated with %s",
	    tskid, shost_addr));
    set_flg(DHCPC_MAIN_FLAGS, DHCP_FLG_DUPLICATED);
    return false;
}
